#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <syslog.h>
#include <sys/types.h>
#include <asm/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <linux/ip.h>
#include <linux/icmp.h>
extern int errno;

#ifndef NOFILE
#define NOFILE 1024
#endif

int go_background(void);
char *hostlookup(unsigned long int);

struct ippkt
{
  struct iphdr ip;
  struct icmphdr icmp;
  char buffer[10000];
} pkt;

char * icmp_unreach[16]=
{
  "Network Unreachable",
  "Host Unreachable",
  "Protocol Unreachable",
  "Port Unreachable",
  "Fragmentation Needed - DF set",
  "Source Route failed",
  "Network Unknown",
  "Host Unknown",
  "Host Isolated",
  "ICMP_NET_ANO",
  "ICMP_HOST_ANO",
  "ICMP_NET_UNR_TOS",
  "ICMP_HOST_UNR_TOS",
  "Packet filtered",
  "Precedence violation",
  "Precedence cut off"
};

char * icmp_redir[4]=
{
  "Net",
  "Host",
  "Net for TOS",
  "Host for TOS"
};

char * icmp_time_exceeded[2]=
{
  "TTL count exceeded",
  "Fragment Reass time exceeded"
};

int go_background(void)
{
  int fd;
  int fs;
  
  if(getppid() != 1)
    {
      signal(SIGTTOU, SIG_IGN);
      signal(SIGTTIN, SIG_IGN);
      signal(SIGTSTP, SIG_IGN);
      fs=fork();
      if(fs < 0)
	{
	  perror("icmplog: fork");
	  exit(1);
	}
      if(fs > 0) exit(0);
      setpgrp();
      fd=open("/dev/tty", O_RDWR);
      if(fd >= 0)
	{
	  ioctl(fd, TIOCNOTTY, (char *)NULL);
	  close(fd);
	}
    }
  for(fd=0;fd < NOFILE;fd++) close(fd);
  errno=0;
  chdir("/");
  umask(0);
}

int main(void)
{
  int s;
  int i;
  char host1[1024], host2[1024];
  char hostname[2200];
  char code_txt[25];

  setuid(0);   
  if(geteuid() != 0)
    {
      printf("This program requires root privledges\n");
      exit(0);
    }
  go_background();
  s=socket(AF_INET, SOCK_RAW, 1);
  openlog("icmplog", 0, LOG_DAEMON);
  
  while(1)
    {
      read(s, (struct ippkt *)&pkt, 9999);
      if (pkt.ip.saddr != pkt.ip.daddr)
#if 1
        {
          strcpy(host1, hostlookup(pkt.ip.saddr));
          strcpy(host2, hostlookup(pkt.ip.daddr));
          sprintf(hostname,"%s -> %s", host1, host2);
        }
#else
        sprintf(hostname,"%08X -> %08X",
                pkt.ip.saddr,
                pkt.ip.daddr);
#endif
      else
        sprintf(hostname,"%s", hostlookup(pkt.ip.saddr));
      if (pkt.icmp.code != 0)
        sprintf(code_txt," (code %d)", pkt.icmp.code);
      else
        sprintf(code_txt,"");
      if(pkt.ip.ihl != 5)
	syslog(LOG_NOTICE, "suspicious ip options from %s (header size %d ??)",
	       hostlookup(pkt.ip.saddr),
	       pkt.ip.ihl);
      else if(pkt.icmp.type == ICMP_ECHOREPLY)
	syslog(LOG_NOTICE,"%s: ping reply%s (id:%d, seq: %d)",
	       hostname, code_txt,
               pkt.icmp.un.echo.id, pkt.icmp.un.echo.sequence);
      else if(pkt.icmp.type == ICMP_DEST_UNREACH)
	syslog(LOG_NOTICE,"%s: %s (%08X)",
	       hostname, icmp_unreach[pkt.icmp.code],
               pkt.icmp.un.gateway);
      else if(pkt.icmp.type == ICMP_SOURCE_QUENCH)
	syslog(LOG_NOTICE,"%s: source quench%s",
	       hostname, code_txt);
      else if(pkt.icmp.type == ICMP_REDIRECT)
	syslog(LOG_NOTICE,"%s: redirect %s %s",
	       hostname, icmp_redir[pkt.icmp.code],
               hostlookup(pkt.icmp.un.gateway));
      else if(pkt.icmp.type == ICMP_ECHO)
	syslog(LOG_NOTICE,"%s: ping%s (id:%d, seq: %d)",
	       hostname, code_txt,
               pkt.icmp.un.echo.id, pkt.icmp.un.echo.sequence);
      else if(pkt.icmp.type == ICMP_TIME_EXCEEDED)
	syslog(LOG_NOTICE,"%s: %s",
	       hostname, icmp_time_exceeded[pkt.icmp.code]);
      else if(pkt.icmp.type == ICMP_PARAMETERPROB)
	syslog(LOG_NOTICE,"%s: parameter problem%s (%08X)",
	       hostname, code_txt,
               pkt.icmp.un.gateway);
      else if(pkt.icmp.type == ICMP_TIMESTAMP)
	syslog(LOG_NOTICE,"%s: timestamp request%s",
	       hostname, code_txt);
      else if(pkt.icmp.type == ICMP_TIMESTAMPREPLY)
	syslog(LOG_NOTICE,"%s: timestamp reply%s",
	       hostname, code_txt);
      else if(pkt.icmp.type == ICMP_INFO_REQUEST)
	syslog(LOG_NOTICE,"%s: info request%s (%08X)",
	       hostname, code_txt,
               pkt.icmp.un.gateway);
      else if(pkt.icmp.type == ICMP_INFO_REPLY)
	syslog(LOG_NOTICE,"%s: info reply%s (%08X)",
	       hostname, code_txt,
               pkt.icmp.un.gateway);
      else if(pkt.icmp.type == ICMP_ADDRESS)
	syslog(LOG_NOTICE,"%s: address mask request%s (%08X)",
	       hostname, code_txt,
               pkt.icmp.un.gateway);
      else if(pkt.icmp.type == ICMP_ADDRESSREPLY)
	syslog(LOG_NOTICE,"%s: address mask reply%s (%08X)",
	       hostname, code_txt,
               pkt.icmp.un.gateway);
      else syslog(LOG_NOTICE,"%s: type %d, code %d (%08X)",
	       hostname,
	       pkt.icmp.type,
	       pkt.icmp.code,
	       pkt.icmp.un.gateway);
    }
}


char *hostlookup(unsigned long int in)
{
  struct in_addr i;
  struct hostent *he;
  
  i.s_addr=in;
  he=gethostbyaddr((char *)&i, sizeof(struct in_addr),AF_INET);
  if(he == NULL) return inet_ntoa(i);
  else return he->h_name;
}
